/*
  $Id: Join.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef _Join_
#define _Join_

#include "Skyline.h"

#include <fstream>
#include <iomanip>

#include <sys/times.h>

using namespace std;

//-- Join

template <class TupleR, class TupleS> 
float join(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
	  const HashIdxUnsigned &indexS)
{
  // timing
  clock_t start, ends;

  //-------
  //-- JOIN
  //-------
  VectPairUnsigned *joinRS = new VectPairUnsigned();
  start = clock();
  for (unsigned iR = 0; iR < datasetR.size(); iR++) 
    {
      if (datasetR[iR].attrA == TupleR::qattrA
#ifdef DIM_2_X
	  && datasetR[iR].attrB == TupleR::qattrB
#endif
	  ) 
	{
	  PairHashIdxUnsignedCIt idxS = indexS.equal_range(datasetR[iR].attrJ);
	  for (HashIdxUnsignedCIt it_idxS = idxS.first; it_idxS != idxS.second; ++it_idxS)
	    {
	      if (datasetS[it_idxS->second].attrA == TupleS::qattrA 
#ifdef DIM_X_2
		  && datasetS[it_idxS->second].attrB == TupleS::qattrB
#endif
		  )
		{
		  joinRS->push_back(PairUnsigned(iR, it_idxS->second));
		}
	    }
	}
      
    }
  ends = clock();
  float time = static_cast<float>(ends - start) / CLOCKS_PER_SEC;

#ifdef PRINT_INFO
  cout << "join running time\t" << time << endl;
  cerr << "size\t" << joinRS->size() << endl;
#endif

#ifdef PRINT_SKYLINE
  for (VectPairUnsignedCIt it = joinRS->begin(); it != joinRS->end(); ++it)
  {
    cout << datasetR[it->first] << "\t|\t" << datasetS[it->second] << endl;
  }
#endif

  delete joinRS;

  return time;
}

//-- Join, Skyline

template <class TupleR, class TupleS> 
float skylineJ(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
	       const HashIdxUnsigned &indexS, 
	       unsigned &sky_size)
{
  // timing
  clock_t start, ends;

  //-------
  //-- JOIN, SKYLINE
  //-------
  VectPairUnsigned *joinRS = new VectPairUnsigned();
  VectPairUnsigned *globalSky = new VectPairUnsigned();
  //-- JOIN
  start = clock();
  for (unsigned iR = 0; iR < datasetR.size(); iR++) 
    {
      PairHashIdxUnsignedCIt idxS = indexS.equal_range(datasetR[iR].attrJ);
      for (HashIdxUnsignedCIt it_idxS = idxS.first; it_idxS != idxS.second; ++it_idxS)
	{
	  joinRS->push_back(PairUnsigned(iR, it_idxS->second));
	}
    }
  //-- SKYLINE
  globalSkyline(datasetR, datasetS, *joinRS, *globalSky);
  ends = clock();
  float time = static_cast<float>(ends - start) / CLOCKS_PER_SEC;
  sky_size = globalSky->size();

#ifdef PRINT_INFO
  cout << "skyline +J running time\t" << time << endl;
  cerr << "sizej\t" << joinRS->size() << endl;
  cerr << "size\t" << globalSky->size() << endl;
#endif

#ifdef PRINT_SKYLINE
  for (VectPairUnsignedCIt it = globalSky->begin(); it != globalSky->end(); ++it)
  {
    cout << datasetR[it->first] << "\t|\t" << datasetS[it->second] << endl;
  }
#endif

  delete joinRS;
  delete globalSky;

  return time;
}

//-- Skyline

template <class TupleR, class TupleS> 
float skyline(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
	      const HashIdxUnsigned &indexS, 
	      unsigned &sky_size)
{
  // timing
  clock_t start, ends;

  //----------
  //-- SKILINE
  //----------
  VectPairUnsigned *globalSky = new VectPairUnsigned();
  VectUnsigned *localSky = new VectUnsigned();

  //-- SKYLINE - without Local Skyline
  start = clock();
  for (unsigned iR = 0; iR < datasetR.size(); iR++) 
    {
      PairHashIdxUnsignedCIt idxS = indexS.equal_range(datasetR[iR].attrJ);
      globalSkyline(datasetR, datasetS, iR, idxS, *globalSky);
    }
  ends = clock();
  float time = static_cast<float>(ends - start) / CLOCKS_PER_SEC;
  sky_size = globalSky->size();

#ifdef PRINT_INFO
  cout << "skyline -L running time\t" << time << endl;
  cerr << "size\t" << globalSky->size() << endl;
#endif

#ifdef PRINT_SKYLINE
  for (VectPairUnsignedCIt it = globalSky->begin(); it != globalSky->end(); ++it)
  {
    cout << datasetR[it->first] << "\t|\t" << datasetS[it->second] << endl;
  }
#endif

  delete localSky;
  delete globalSky;

  return time;
}

//-- Skyline L

template <class TupleR, class TupleS> 
float skylineL(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
	       const HashIdxUnsigned &indexS, 
	       unsigned &sky_size)
{
  // timing
  clock_t start, ends;

  //----------
  //-- SKILINE
  //----------
  VectPairUnsigned *globalSky = new VectPairUnsigned();
  VectUnsigned *localSky = new VectUnsigned();

  //-- SKYLINE - with Local Skyline
  start = clock();
  for (unsigned iR = 0; iR < datasetR.size(); iR++) 
    {
      PairHashIdxUnsignedCIt idxS = indexS.equal_range(datasetR[iR].attrJ);
      localSkyline(datasetS, idxS, *localSky);
      globalSkyline(datasetR, datasetS, iR, *localSky, *globalSky);
      localSky->clear();
    }
  ends = clock();
  float time = static_cast<float>(ends - start) / CLOCKS_PER_SEC;
  sky_size = globalSky->size();

#ifdef PRINT_INFO
  cout << "skyline +L running time\t" << time << endl;
  cerr << "size\t" << globalSky->size() << endl;
#endif

#ifdef PRINT_SKYLINE
  for (VectPairUnsignedCIt it = globalSky->begin(); it != globalSky->end(); ++it)
  {
    cout << datasetR[it->first] << "\t|\t" << datasetS[it->second] << endl;
  }
#endif

  delete localSky;
  delete globalSky;
  
  return time;
}

//-- Sorted-Access

template <class TupleR, class TupleS> 
float sorted(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
	     const HashIdxUnsigned &indexR, const HashIdxUnsigned &indexS, 
	     unsigned &sky_size)
{
  // timing
  clock_t start, ends;

  //----------
  //-- SKILINE
  //----------

  //-- create "sorted lists"
  VectUnsigned *srtLst[dim];
  unsigned crt_lst = 0;
  srtLst[crt_lst] = new VectUnsigned();
  srtLst[crt_lst]->reserve(datasetR.size());
  Inc inc;
  generate_n(back_inserter(*srtLst[crt_lst]), datasetR.size(), inc);
  crt_lst++;

#ifdef DIM_2_X
  srtLst[crt_lst] = new VectUnsigned(*srtLst[crt_lst - 1]);
  crt_lst++;
#endif

#ifdef DIM_3_X
  srtLst[crt_lst] = new VectUnsigned(*srtLst[crt_lst - 1]);
  crt_lst++;
#endif

  srtLst[crt_lst] = new VectUnsigned();
  srtLst[crt_lst]->reserve(datasetS.size());
  inc.reset();
  generate_n(back_inserter(*srtLst[crt_lst]), datasetS.size(), inc);
  crt_lst++;

#ifdef DIM_X_2
  srtLst[crt_lst] = new VectUnsigned(*srtLst[crt_lst - 1]);
  crt_lst++;
#endif

#ifdef DIM_X_3
  srtLst[crt_lst] = new VectUnsigned(*srtLst[crt_lst - 1]);
  crt_lst++;
#endif

  //-- sort "sorted lists"
  crt_lst = 0;
  SortPredA<TupleR> relR_spA(datasetR);
  sort(srtLst[crt_lst]->begin(), srtLst[crt_lst]->end(), relR_spA);
  crt_lst++;

#ifdef DIM_2_X
  SortPredB<TupleR> relR_spB(datasetR);
  sort(srtLst[crt_lst]->begin(), srtLst[crt_lst]->end(), relR_spB);
  crt_lst++;
#endif

#ifdef DIM_3_X
  SortPredC<TupleR> relR_spC(datasetR);
  sort(srtLst[crt_lst]->begin(), srtLst[crt_lst]->end(), relR_spC);
  crt_lst++;
#endif

  SortPredA<TupleS> relS_spA(datasetS);
  sort(srtLst[crt_lst]->begin(), srtLst[crt_lst]->end(), relS_spA);
  crt_lst++;
  
#ifdef DIM_X_2
  SortPredB<TupleS> relS_spB(datasetS);
  sort(srtLst[crt_lst]->begin(), srtLst[crt_lst]->end(), relS_spB);
  crt_lst++;
#endif

#ifdef DIM_X_3
  SortPredC<TupleS> relS_spC(datasetS);
  sort(srtLst[crt_lst]->begin(), srtLst[crt_lst]->end(), relS_spC);
  crt_lst++;
#endif

  SetPairUnsigned
    *seen = new SetPairUnsigned();
  VectPairUnsigned
    *local_seen = new VectPairUnsigned(), 
    *globalSky = new VectPairUnsigned();
  VectUnsignedCIt srtLstIt[dim];
  for (unsigned i = 0; i < dim; i++)
    {
      srtLstIt[i] = srtLst[i]->begin();
    }

  int lst_idx = 0;
  bool found_frontiere = false;
  //-- FRONTIERE
  start = clock();
  while (!found_frontiere) 
    {
      // (id, lst_idx) represent a value for an attribute
      unsigned id;
      bool
	good_id = false, 
	no_good_id = false;
      int old_lst_idx = lst_idx;
      //-- if current list is depleted, take next list
      while (!good_id) 
	{
	  if (srtLstIt[lst_idx] != srtLst[lst_idx]->end()) 
	    {
	      id = *srtLstIt[lst_idx];
	      ++srtLstIt[lst_idx];
	      good_id = true;
	    }
	  else
	    {
	      lst_idx = (lst_idx + 1) % dim;
	      if (lst_idx == old_lst_idx) 
		{
		  no_good_id = true;
		  break;
		}
	    }
	}
      if (no_good_id) 
	{
	  break;
	}

      // retrieve that tuple and all the tuples that join with it
      local_seen->clear();
      if (lst_idx < DIM_S_A) 
	{ // attribute is from R
	  PairHashIdxUnsignedCIt idxS = indexS.equal_range(datasetR[id].attrJ);
	  for (HashIdxUnsignedCIt it_idxS = idxS.first; it_idxS != idxS.second; ++it_idxS)
	    {
	      local_seen->push_back(PairUnsigned(id, it_idxS->second));
	    }
	}
      else 
	{ // attribute is from S
	  PairHashIdxUnsignedCIt idxR = indexR.equal_range(datasetS[id].attrJ);
	  for (HashIdxUnsignedCIt it_idxR = idxR.first; it_idxR != idxR.second; ++it_idxR)
	    {
	      local_seen->push_back(PairUnsigned(it_idxR->second, id));
	    }
	}

      copy(local_seen->begin(), local_seen->end(), inserter(*seen, seen->begin()));

      for (VectPairUnsignedCIt it = local_seen->begin(); it != local_seen->end(); ++it) 
	{
	  if ((lst_idx == 0 || 
	       srtLstIt[0] == srtLst[0]->end() || 
	       datasetR[*srtLstIt[0]].relaxA() > datasetR[it->first].relaxA())
#ifdef DIM_2_X
	      &&
	      (lst_idx == 1 || 
	       srtLstIt[1] == srtLst[1]->end() || 
	       datasetR[*srtLstIt[1]].relaxB() > datasetR[it->first].relaxB())
#endif
#ifdef DIM_3_X
	      &&
	      (lst_idx == 2 || 
	       srtLstIt[2] == srtLst[2]->end() || 
	       datasetR[*srtLstIt[2]].relaxC() > datasetR[it->first].relaxC())
#endif
	      && 
	      (lst_idx == DIM_S_A || 
	       srtLstIt[DIM_S_A] == srtLst[DIM_S_A]->end() || 
	       datasetS[*srtLstIt[DIM_S_A]].relaxA() > datasetS[it->second].relaxA())
#ifdef DIM_X_2
	      &&
	      (lst_idx == DIM_S_A + 1 || 
	       srtLstIt[DIM_S_A + 1] == srtLst[DIM_S_A + 1]->end() || 
	       datasetS[*srtLstIt[DIM_S_A + 1]].relaxB() > datasetS[it->second].relaxB())
#endif
#ifdef DIM_X_3
	      &&
	      (lst_idx == DIM_S_A + 2 || 
	       srtLstIt[DIM_S_A + 2] == srtLst[DIM_S_A + 2]->end() || 
	       datasetS[*srtLstIt[DIM_S_A + 2]].relaxC() > datasetS[it->second].relaxC())
#endif
	      )
	    {
	      found_frontiere = true;
	      break;
	    }
	}

      lst_idx = (lst_idx + 1) % dim;
    }

  //-- SKYLINE
  globalSkyline(datasetR, datasetS, *seen, *globalSky);
  ends = clock();
  float time = static_cast<float>(ends - start) / CLOCKS_PER_SEC;
  sky_size = globalSky->size();

#ifdef PRINT_INFO
  cout << "skyline +S running time\t" << time << endl;
  cerr << "sizef\t" << seen->size() << endl;
  cerr << "size\t" << globalSky->size() << endl;
#endif

#ifdef PRINT_SKYLINE
  for (VectPairUnsignedCIt it = globalSky->begin(); it != globalSky->end(); ++it)
  {
    cout << datasetR[it->first] << "\t|\t" << datasetS[it->second] << endl;
  }  
#endif

  delete seen;
  delete local_seen;
  delete globalSky;

  for (unsigned i = 0; i < dim; i++) 
    {
      delete srtLst[i];
    }
  
  return time;
}

//--

#include <numeric>

using namespace std;

//--

template <class Container> 
inline float average(const Container& vect) 
{
  return accumulate(vect.begin(), vect.end(), 0.) / vect.size();
}

typedef vector<float> VectFloat;

//-- run

template <class TupleR, class TupleS> 
void run(const vector<TupleR> &datasetR, const vector<TupleS> &datasetS, 
	 const HashIdxUnsigned &indexR, const HashIdxUnsigned &indexS, 
	 const unsigned no_runs = 1, const unsigned size = 0, const unsigned card = 0)
{
  VectFloat times_join, times_skyline[4];

  VectUnsigned sizes_skyline[4];

  for (unsigned i = 0; i < sizeof(TupleR::tupleR_qattrA) / sizeof(TupleR::qattrA); i++) 
    {
#ifdef DIM_2_X
      for (unsigned j = 0; j < sizeof(TupleR::tupleR_qattrB) / sizeof(TupleR::qattrB); j++)
#endif
	{
	  for (unsigned k = 0; k < sizeof(TupleS::tupleS_qattrA) / sizeof(TupleS::qattrA); k++) 
	    {
#ifdef DIM_X_2
	      for (unsigned l = 0; l < sizeof(TupleS::tupleS_qattrB) / sizeof(TupleS::qattrB); l++) 
#endif
		{	
#ifdef PRINT_INFO
		  cout << "query\t" << i << k << endl;
#endif

#ifdef PRINT_SKYLINE
                  cout << "query R: " << TupleR::tupleR_qattrA[i]
#ifdef DIM_X_2
                       << ", " << TupleR::tupleR_qattrB[j]
#endif
                       << ", S: " << TupleS::tupleS_qattrA[k]
#ifdef DIM_X_2
                       << ", " << TupleS::tupleS_qattrB[l]
#endif
                       << endl;
#endif

		  TupleR::qattrA = TupleR::tupleR_qattrA[i];

#ifdef DIM_2_X
		  TupleR::qattrB = TupleR::tupleR_qattrB[j];
#endif

#ifdef DIM_3_X
		  TupleR::qattrC = TupleR::tupleR_qattrC[0];
#endif

		  TupleS::qattrA = TupleS::tupleS_qattrA[k];

#ifdef DIM_X_2
		  TupleS::qattrB = TupleS::tupleS_qattrB[l];
#endif

#ifdef DIM_X_3
		  TupleS::qattrC = TupleS::tupleS_qattrC[0];
#endif

		  for (unsigned m = 0; m < no_runs; m++) 
		    {
		      // cout << '.';
		      // cout.flush();
		      
		      // times_join.push_back(join(datasetR, datasetS, indexS));
		      
		      unsigned size_skyline;
		      
		      times_skyline[0].push_back(skylineJ(datasetR, datasetS, indexS, size_skyline));
		      sizes_skyline[0].push_back(size_skyline);
		      
		      // times_skyline[1].push_back(skyline(datasetR, datasetS, indexS, size_skyline));
		      // sizes_skyline[1].push_back(size_skyline);

		      // times_skyline[2].push_back(skylineL(datasetR, datasetS, indexS, size_skyline));		  
		      // sizes_skyline[2].push_back(size_skyline);

 		      // times_skyline[3].push_back(sorted(datasetR, datasetS, indexR, indexS, size_skyline));
		      // sizes_skyline[3].push_back(size_skyline);
		    }
		}
	    }
	}
    }
  cout << endl;

#ifdef PRINT_INFO
  cout << fixed << setprecision(2);
  // cout << "---" << endl;
  // cout << "no queries\t" << times_join.size() << endl;
  // cout << "avg join\t" << average(times_join) << endl;
  // cout << "avg skyline +J\t" << average(times_skyline[0]) << endl;
  // cout << "avg skyline -L\t" << average(times_skyline[1]) << endl;
  // cout << "avg skyline +L\t" << average(times_skyline[2]) << endl;
  // cout << "avg skyline +S\t" << average(times_skyline[3]) << endl;
#endif

#ifdef WRITE_RESULTS
  ofstream *fout;  
  for (unsigned i = 0; i < sizeof(out_skyline_filename) / sizeof(string); i++) 
    {
      fout = new ofstream(out_skyline_fullpath[i].c_str(), ios::out | ios::app);
      (*fout) << size << " " << average(times_skyline[i]) << endl;
      delete fout;
    }

//   for (unsigned i = 0; i < sizeof(out_skyline_filename) / sizeof(string); i++) 
//     {
//       fout = new ofstream(out_skyline_size_fullpath[i].c_str(), ios::out | ios::app);
//       (*fout) << dim << " " << average(sizes_skyline[i]) << endl;
//       delete fout;
//     }
#endif
}

//--

#endif
